/*
 * ID3v1 header parser
 * Copyright (c) 2003 Fabrice Bellard
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "id3v1.h"
#include "libavcodec/avcodec.h"

#ifdef _MSC_VER
const char *ff_id3v1_genre_str[ID3v1_GENRE_MAX + 1] = {
 "Blues",              /*   1 (000000) */
 "Classic Rock",       /*   2 (000001) */
 "Country",            /*   3 (000002) */
 "Dance",              /*   4 (000003) */
 "Disco",              /*   5 (000004) */
 "Funk",               /*   6 (000005) */
 "Grunge",             /*   7 (000006) */
 "Hip-Hop",            /*   8 (000007) */
 "Jazz",               /*   9 (000008) */
 "Metal",              /*  10 (000009) */
 "New Age",            /*  11 (000010) */
 "Oldies",             /*  12 (000011) */
 "Other",              /*  13 (000012) */
 "Pop",                /*  14 (000013) */
 "R&B",                /*  15 (000014) */
 "Rap",                /*  16 (000015) */
 "Reggae",             /*  17 (000016) */
 "Rock",               /*  18 (000017) */
 "Techno",             /*  19 (000018) */
 "Industrial",         /*  20 (000019) */
 "Alternative",        /*  21 (000020) */
 "Ska",                /*  22 (000021) */
 "Death Metal",        /*  23 (000022) */
 "Pranks",             /*  24 (000023) */
 "Soundtrack",         /*  25 (000024) */
 "Euro-Techno",        /*  26 (000025) */
 "Ambient",            /*  27 (000026) */
 "Trip-Hop",           /*  28 (000027) */
 "Vocal",              /*  29 (000028) */
 "Jazz+Funk",          /*  30 (000029) */
 "Fusion",             /*  31 (000030) */
 "Trance",             /*  32 (000031) */
 "Classical",          /*  33 (000032) */
 "Instrumental",       /*  34 (000033) */
 "Acid",               /*  35 (000034) */
 "House",              /*  36 (000035) */
 "Game",               /*  37 (000036) */
 "Sound Clip",         /*  38 (000037) */
 "Gospel",             /*  39 (000038) */
 "Noise",              /*  40 (000039) */
 "AlternRock",         /*  41 (000040) */
 "Bass",               /*  42 (000041) */
 "Soul",               /*  43 (000042) */
 "Punk",               /*  44 (000043) */
 "Space",              /*  45 (000044) */
 "Meditative",         /*  46 (000045) */
 "Instrumental Pop",   /*  47 (000046) */
 "Instrumental Rock",  /*  48 (000047) */
 "Ethnic",             /*  49 (000048) */
 "Gothic",             /*  50 (000049) */
 "Darkwave",           /*  51 (000050) */
 "Techno-Industrial",  /*  52 (000051) */
 "Electronic",         /*  53 (000052) */
 "Pop-Folk",           /*  54 (000053) */
 "Eurodance",          /*  55 (000054) */
 "Dream",              /*  56 (000055) */
 "Southern Rock",      /*  57 (000056) */
 "Comedy",             /*  58 (000057) */
 "Cult",               /*  59 (000058) */
 "Gangsta",            /*  60 (000059) */
 "Top 40",             /*  61 (000060) */
 "Christian Rap",      /*  62 (000061) */
 "Pop/Funk",           /*  63 (000062) */
 "Jungle",             /*  64 (000063) */
 "Native American",    /*  65 (000064) */
 "Cabaret",            /*  66 (000065) */
 "New Wave",           /*  67 (000066) */
 "Psychadelic",        /*  68 (000067) */
 "Rave",               /*  69 (000068) */
 "Showtunes",          /*  70 (000069) */
 "Trailer",            /*  71 (000070) */
 "Lo-Fi",              /*  72 (000071) */
 "Tribal",             /*  73 (000072) */
 "Acid Punk",          /*  74 (000073) */
 "Acid Jazz",          /*  75 (000074) */
 "Polka",              /*  76 (000075) */
 "Retro",              /*  77 (000076) */
 "Musical",            /*  78 (000077) */
 "Rock & Roll",        /*  79 (000078) */
 "Hard Rock",          /*  80 (000079) */
 "Folk",               /*  81 (000080) */
 "Folk-Rock",          /*  82 (000081) */
 "National Folk",      /*  83 (000082) */
 "Swing",              /*  84 (000083) */
 "Fast Fusion",        /*  85 (000084) */
 "Bebob",              /*  86 (000085) */
 "Latin",              /*  87 (000086) */
 "Revival",            /*  88 (000087) */
 "Celtic",             /*  89 (000088) */
 "Bluegrass",          /*  90 (000089) */
 "Avantgarde",         /*  91 (000090) */
 "Gothic Rock",        /*  92 (000091) */
 "Progressive Rock",   /*  93 (000092) */
 "Psychedelic Rock",   /*  94 (000093) */
 "Symphonic Rock",     /*  95 (000094) */
 "Slow Rock",          /*  96 (000095) */
 "Big Band",           /*  97 (000096) */
 "Chorus",             /*  98 (000097) */
 "Easy Listening",     /*  99 (000098) */
 "Acoustic",           /* 100 (000099) */
 "Humour",             /* 101 (000100) */
 "Speech",             /* 102 (000101) */
 "Chanson",            /* 103 (000102) */
 "Opera",              /* 104 (000103) */
 "Chamber Music",      /* 105 (000104) */
 "Sonata",             /* 106 (000105) */
 "Symphony",           /* 107 (000106) */
 "Booty Bass",         /* 108 (000107) */
 "Primus",             /* 109 (000108) */
 "Porn Groove",        /* 110 (000109) */
 "Satire",             /* 111 (000110) */
 "Slow Jam",           /* 112 (000111) */
 "Club",               /* 113 (000112) */
 "Tango",              /* 114 (000113) */
 "Samba",              /* 115 (000114) */
 "Folklore",           /* 116 (000115) */
 "Ballad",             /* 117 (000116) */
 "Power Ballad",       /* 118 (000117) */
 "Rhythmic Soul",      /* 119 (000118) */
 "Freestyle",          /* 120 (000119) */
 "Duet",               /* 121 (000120) */
 "Punk Rock",          /* 122 (000121) */
 "Drum Solo",          /* 123 (000122) */
 "A capella",          /* 124 (000123) */
 "Euro-House",         /* 125 (000124) */
 "Dance Hall"          /* 126 (000125) */
};
#else // !_MSC_VER
const char *ff_id3v1_genre_str[ID3v1_GENRE_MAX + 1] = {
      [0] = "Blues",
      [1] = "Classic Rock",
      [2] = "Country",
      [3] = "Dance",
      [4] = "Disco",
      [5] = "Funk",
      [6] = "Grunge",
      [7] = "Hip-Hop",
      [8] = "Jazz",
      [9] = "Metal",
     [10] = "New Age",
     [11] = "Oldies",
     [12] = "Other",
     [13] = "Pop",
     [14] = "R&B",
     [15] = "Rap",
     [16] = "Reggae",
     [17] = "Rock",
     [18] = "Techno",
     [19] = "Industrial",
     [20] = "Alternative",
     [21] = "Ska",
     [22] = "Death Metal",
     [23] = "Pranks",
     [24] = "Soundtrack",
     [25] = "Euro-Techno",
     [26] = "Ambient",
     [27] = "Trip-Hop",
     [28] = "Vocal",
     [29] = "Jazz+Funk",
     [30] = "Fusion",
     [31] = "Trance",
     [32] = "Classical",
     [33] = "Instrumental",
     [34] = "Acid",
     [35] = "House",
     [36] = "Game",
     [37] = "Sound Clip",
     [38] = "Gospel",
     [39] = "Noise",
     [40] = "AlternRock",
     [41] = "Bass",
     [42] = "Soul",
     [43] = "Punk",
     [44] = "Space",
     [45] = "Meditative",
     [46] = "Instrumental Pop",
     [47] = "Instrumental Rock",
     [48] = "Ethnic",
     [49] = "Gothic",
     [50] = "Darkwave",
     [51] = "Techno-Industrial",
     [52] = "Electronic",
     [53] = "Pop-Folk",
     [54] = "Eurodance",
     [55] = "Dream",
     [56] = "Southern Rock",
     [57] = "Comedy",
     [58] = "Cult",
     [59] = "Gangsta",
     [60] = "Top 40",
     [61] = "Christian Rap",
     [62] = "Pop/Funk",
     [63] = "Jungle",
     [64] = "Native American",
     [65] = "Cabaret",
     [66] = "New Wave",
     [67] = "Psychadelic",
     [68] = "Rave",
     [69] = "Showtunes",
     [70] = "Trailer",
     [71] = "Lo-Fi",
     [72] = "Tribal",
     [73] = "Acid Punk",
     [74] = "Acid Jazz",
     [75] = "Polka",
     [76] = "Retro",
     [77] = "Musical",
     [78] = "Rock & Roll",
     [79] = "Hard Rock",
     [80] = "Folk",
     [81] = "Folk-Rock",
     [82] = "National Folk",
     [83] = "Swing",
     [84] = "Fast Fusion",
     [85] = "Bebob",
     [86] = "Latin",
     [87] = "Revival",
     [88] = "Celtic",
     [89] = "Bluegrass",
     [90] = "Avantgarde",
     [91] = "Gothic Rock",
     [92] = "Progressive Rock",
     [93] = "Psychedelic Rock",
     [94] = "Symphonic Rock",
     [95] = "Slow Rock",
     [96] = "Big Band",
     [97] = "Chorus",
     [98] = "Easy Listening",
     [99] = "Acoustic",
    [100] = "Humour",
    [101] = "Speech",
    [102] = "Chanson",
    [103] = "Opera",
    [104] = "Chamber Music",
    [105] = "Sonata",
    [106] = "Symphony",
    [107] = "Booty Bass",
    [108] = "Primus",
    [109] = "Porn Groove",
    [110] = "Satire",
    [111] = "Slow Jam",
    [112] = "Club",
    [113] = "Tango",
    [114] = "Samba",
    [115] = "Folklore",
    [116] = "Ballad",
    [117] = "Power Ballad",
    [118] = "Rhythmic Soul",
    [119] = "Freestyle",
    [120] = "Duet",
    [121] = "Punk Rock",
    [122] = "Drum Solo",
    [123] = "A capella",
    [124] = "Euro-House",
    [125] = "Dance Hall",
};
#endif // _MSC_VER y/n

static void get_string(AVFormatContext *s, const char *key,
                       const uint8_t *buf, int buf_size)
{
    int i, c;
    char *q, str[512];

    q = str;
    for(i = 0; i < buf_size; i++) {
        c = buf[i];
        if (c == '\0')
            break;
        if ((q - str) >= sizeof(str) - 1)
            break;
        *q++ = c;
    }
    *q = '\0';

    if (*str)
        av_metadata_set(&s->metadata, key, str);
}

/**
 * Parse an ID3v1 tag
 *
 * @param buf ID3v1_TAG_SIZE long buffer containing the tag
 */
static int parse_tag(AVFormatContext *s, const uint8_t *buf)
{
    char str[5];
    int genre;

    if (!(buf[0] == 'T' &&
          buf[1] == 'A' &&
          buf[2] == 'G'))
        return -1;
    get_string(s, "title",   buf +  3, 30);
    get_string(s, "author",  buf + 33, 30);
    get_string(s, "album",   buf + 63, 30);
    get_string(s, "year",    buf + 93,  4);
    get_string(s, "comment", buf + 97, 30);
    if (buf[125] == 0 && buf[126] != 0) {
        snprintf(str, sizeof(str), "%d", buf[126]);
        av_metadata_set(&s->metadata, "track", str);
    }
    genre = buf[127];
    if (genre <= ID3v1_GENRE_MAX)
        av_metadata_set(&s->metadata, "genre", ff_id3v1_genre_str[genre]);
    return 0;
}

void ff_id3v1_read(AVFormatContext *s)
{
    int ret, filesize;
    uint8_t buf[ID3v1_TAG_SIZE];

    if (!url_is_streamed(s->pb)) {
        /* XXX: change that */
        filesize = url_fsize(s->pb);
        if (filesize > 128) {
            url_fseek(s->pb, filesize - 128, SEEK_SET);
            ret = get_buffer(s->pb, buf, ID3v1_TAG_SIZE);
            if (ret == ID3v1_TAG_SIZE) {
                parse_tag(s, buf);
            }
            url_fseek(s->pb, 0, SEEK_SET);
        }
    }
}
